@@ -19,6 +19,65 @@
 
 typedef gboolean po_filter_func (PoEntry *);
 
+void
+msgstrx_free(MsgStrX *msgstrx)
+{
+	g_free (msgstrx->str);
+	g_free (msgstrx);
+}
+
+void
+po_list_str_dup(GSList *list)
+{
+	GSList *l;
+	for (l = list; l != NULL; l = l->next) {
+		char *s = l->data;
+		l->data = g_strdup (s);
+	}
+}
+
+void
+po_list_msgstrx_dup(GSList *list)
+{
+	GSList *l;
+	for (l = list; l != NULL; l = l->next) {
+		MsgStrX *s = l->data;
+		MsgStrX *n = g_new (MsgStrX, 1);
+		n->n = s->n;
+		n->str = g_strdup (s->str);
+		l->data = n;
+	}
+}
+
+PoEntry *
+po_entry_copy (PoEntry *ret, PoEntry *po)
+{
+	if (ret == NULL)
+		ret = g_new (PoEntry, 1);
+
+	ret->comments.std = g_slist_copy (po->comments.std);
+	po_list_str_dup(ret->comments.std);
+	ret->comments.pos = g_slist_copy (po->comments.pos);
+	po_list_str_dup(ret->comments.pos);
+	ret->comments.res = g_slist_copy (po->comments.res);
+	po_list_str_dup(ret->comments.res);
+	ret->comments.spec = g_slist_copy (po->comments.spec);
+	po_list_str_dup(ret->comments.spec);
+
+	ret->is_fuzzy = po->is_fuzzy;
+	ret->is_c_format = po->is_c_format;
+
+	ret->ctx = g_strdup (po->ctx);
+	ret->id = g_strdup (po->id);
+	ret->id_plural = g_strdup (po->id_plural);
+	ret->str = g_strdup (po->str);
+
+	ret->msgstrxs = g_slist_copy (po->msgstrxs);
+	po_list_msgstrx_dup(ret->msgstrxs);
+
+	return ret;
+}
+
 void
 po_entry_free (PoEntry *po)
 {
@@ -26,8 +85,12 @@ po_entry_free (PoEntry *po)
 	g_slist_free_custom (po->comments.pos, g_free);
 	g_slist_free_custom (po->comments.res, g_free);
 	g_slist_free_custom (po->comments.spec, g_free);
+	g_slist_free_custom (po->msgstrxs, msgstrx_free);
 	g_free (po->id);
+	g_free (po->id_plural);
+	g_free (po->ctx);
 	g_free (po->str);
+	g_free (po);
 }
 
 void
@@ -37,22 +100,57 @@ po_obsolete_entry_free (PoObsoleteEntry *po)
 	g_slist_free_custom (po->comments.pos, g_free);
 	g_slist_free_custom (po->comments.res, g_free);
 	g_slist_free_custom (po->comments.spec, g_free);
+	g_slist_free_custom (po->msgstrxs, msgstrx_free);
 	g_free (po->id);
+	g_free (po->id_plural);
+	g_free (po->ctx);
 	g_free (po->str);
+	g_free (po);
+}
+
+void
+po_free (PoFile *pof)
+{
+	g_slist_free_custom (pof->entries, po_entry_free);
+	g_slist_free_custom (pof->obsolete_entries, po_obsolete_entry_free);
+	g_free (pof);
 }
 
 /* --- PoEntry filters --- */
 
+gint
+msgstrx_is_same_firstchar (gconstpointer a, gconstpointer b)
+{
+	const MsgStrX *as = a, *bs = b;
+	return ! (as == bs || (as && bs && ((as->str == bs->str) || (as->str && bs->str && as->str[0] == bs->str[0]))));
+}
+
 static gboolean
 po_filter_translated (PoEntry *po)
 {
-	return po->str[0] != '\0';
+	if (po->str)
+		return po->str[0] != '\0';
+	
+	/* With plural forms, only return true if ALL forms are translated. The
+	 * list is guaranteed to be non-empty by the grammar */
+	else {
+		MsgStrX empty = { "", 0 };
+		return NULL == g_slist_find_custom (po->msgstrxs, &empty, msgstrx_is_same_firstchar);
+	}
 }
 
 static gboolean
 po_filter_not_translated (PoEntry *po)
 {
-	return po->str[0] == '\0';
+	if (po->str)
+		return po->str[0] == '\0';
+
+	/* With plural forms, only return true if ANY forms are not translated.
+	 * The list is guaranteed to be non-empty by the grammar */
+	else {
+		MsgStrX empty = { "", 0 };
+		return NULL != g_slist_find_custom (po->msgstrxs, &empty, msgstrx_is_same_firstchar);
+	}
 }
 
 static gboolean
@@ -77,6 +175,8 @@ po_apply_filter (PoFile *pof, po_filter_func *filter)
 	for (npo_list = NULL, l = pof->entries; l != NULL; l = l->next) {
 		if (filter ((PoEntry *) l->data)) {
 			npo_list = g_slist_prepend (npo_list, l->data);
+		} else {
+			po_entry_free (l->data);
 		}
 	}
 	g_slist_free (pof->entries);
@@ -126,8 +226,16 @@ po_copy_msgid (PoFile *pof)
 	for (l = pof->entries; l != NULL; l = l->next) {
 		PoEntry *po = l->data;
 
-		g_free (po->str);
-		po->str = g_strdup (po->id);
+		if (po->str) {
+			g_free (po->str);
+			po->str = g_strdup (po->id);
+		} else {
+			MsgStrX *m = g_new (MsgStrX, 1);
+			m->n = 0;
+			m->str = g_strdup (po->id);
+			g_slist_free_custom (po->msgstrxs, msgstrx_free);
+			po->msgstrxs = g_slist_append(NULL, m);
+		}
 	}
 
 }
@@ -135,14 +243,15 @@ po_copy_msgid (PoFile *pof)
 /* --- */
 
 typedef enum {
-	NO_ID           = 1 << 0,
-	NO_STR          = 1 << 1,
-	NO_STD_COMMENT  = 1 << 2,
-	NO_POS_COMMENT  = 1 << 3,
-	NO_SPEC_COMMENT = 1 << 4,
-	NO_RES_COMMENT  = 1 << 5,
-	NO_TRANSLATION  = 1 << 6,
-	NO_LINF         = 1 << 7
+	NO_CTX		= 1 << 0,
+	NO_ID           = 1 << 1,
+	NO_STR          = 1 << 2,
+	NO_STD_COMMENT  = 1 << 3,
+	NO_POS_COMMENT  = 1 << 4,
+	NO_SPEC_COMMENT = 1 << 5,
+	NO_RES_COMMENT  = 1 << 6,
+	NO_TRANSLATION  = 1 << 7,
+	NO_LINF         = 1 << 8
 } po_write_modes;
 
 enum {
@@ -195,7 +304,7 @@ print_multi_line (const char *s, int start_offset, const char *prefix)
 				word_len++;
 
 			if (line_has_eol && cur[word_len] == '\0' &&
-			    cur[word_len - 1] != SEP1 && cur[word_len - 1] != SEP2) {
+			    (word_len == 0 || (cur[word_len - 1] != SEP1 && cur[word_len - 1] != SEP2))) {
 				eol_len = 2;
 			} else {
 				eol_len = 0;
@@ -221,6 +330,28 @@ print_multi_line (const char *s, int start_offset, const char *prefix)
 	g_strfreev (lines);
 }
 
+static void
+write_msgstr (char *prefix, char *str, GSList *strn, po_write_modes mode)
+{
+	int prefix_len = strlen(prefix);
+
+	if (!(mode & NO_TRANSLATION)) {
+		if (str) {
+			printf ("%smsgstr ", prefix);
+			print_multi_line (str, 7 + prefix_len, prefix);
+		} else {
+			GSList *x;
+			for (x = strn; x != NULL; x = x->next) {
+				MsgStrX *m = x->data;
+				printf ("%smsgstr[%d] ", prefix, m->n);
+				print_multi_line (m->str, 10 + prefix_len, prefix);
+			}
+		}
+	} else {
+		printf ("%smsgstr \"\"\n", prefix);
+	}
+}
+
 static void
 po_write (PoFile *pof, po_write_modes mode)
 {
@@ -272,17 +403,20 @@ po_write (PoFile *pof, po_write_modes mode)
 				printf ("#,%s\n", (char *) ll->data);
 			}
 		}
+		if ((!(mode & NO_CTX)) && po->ctx) {
+			printf ("msgctxt ");
+			print_multi_line (po->ctx, 8, "");
+		}
 		if (!(mode & NO_ID)) {
 			printf ("msgid ");
 			print_multi_line (po->id, 6, "");
+			if (po->id_plural) {
+				printf ("msgid_plural ");
+				print_multi_line (po->id_plural, 13, "");
+			}
 		}
 		if (!(mode & NO_STR)) {
-			if (!(mode & NO_TRANSLATION)) {
-				printf ("msgstr ");
-				print_multi_line (po->str, 7, "");
-			} else {
-				printf ("msgstr \"\"\n");
-			}
+			write_msgstr ("", po->str, po->msgstrxs, mode);
 		}
 
 		if (l->next != NULL) {
@@ -298,23 +432,32 @@ po_write (PoFile *pof, po_write_modes mode)
 		PoObsoleteEntry *po = l->data;
 		GSList *ll;
 
+		if (!(mode & NO_STD_COMMENT)) {
+			for (ll = po->comments.std; ll != NULL; ll = ll->next) {
+				printf ("#%s\n", (char *) ll->data);
+			}
+		}
 		if (!(mode & NO_SPEC_COMMENT)) {
 			for (ll = po->comments.spec; ll != NULL; ll = ll->next) {
 				printf ("#,%s\n", (char *) ll->data);
 			}
 		}
 
+		if ((!(mode & NO_CTX)) && po->ctx) {
+			printf ("#~ msgctxt ");
+			print_multi_line (po->ctx, 11, "#~ ");
+		}
+
 		if (!(mode & NO_ID)) {
 			printf ("#~ msgid ");
 			print_multi_line (po->id, 9, "#~ ");
+			if (po->id_plural) {
+				printf ("#~ msgid_plural ");
+				print_multi_line (po->id_plural, 16, "#~ ");
+			}
 		}
 		if (!(mode & NO_STR)) {
-			if (!(mode & NO_TRANSLATION)) {
-				printf ("#~ msgstr ");
-				print_multi_line (po->str, 10, "#~ ");
-			} else {
-				printf ("#~ msgstr \"\"\n");
-			}
+			write_msgstr ("#~ ", po->str, po->msgstrxs, mode);
 		}
 
 		if (l->next != NULL) {
@@ -349,13 +492,8 @@ po_set_update (PoEntry_set *po_set, GSList *po_list)
 		PoEntry *po = (PoEntry *) l->data, *hpo;
 
 		if ((hpo = g_hash_table_lookup (po_set, po->id)) != NULL) {
-			hpo->str = po->str;
-			hpo->comments.std = po->comments.std;
-			hpo->comments.pos = po->comments.pos;
-			hpo->comments.res = po->comments.res;
-			hpo->comments.spec = po->comments.spec;
-			hpo->is_fuzzy = po->is_fuzzy;
-			hpo->is_c_format = po->is_c_format;
+			/* making a deep copy, since we are about to free po_list */
+			po_entry_copy (hpo, po);
 		} else {
 			g_warning (_("Unknown msgid: %s"), po->id);
 		}
@@ -377,7 +515,7 @@ main (int argc, char **argv)
 		FILENAME1 [FILENAME2]
 		[-f f|nf|t|nt]
 		[-s] [-c]
-		[-n id|str|cmt|ucmt|pcmt|scmt|dcmt|tr|linf]...
+		[-n ctxt|id|str|cmt|ucmt|pcmt|scmt|dcmt|tr|linf]...
 		[-h]
 	 */
 	while ((c = getopt (argc, argv, "f:n:sch")) != EOF) {
@@ -390,7 +528,9 @@ main (int argc, char **argv)
 				exit (0);
 				break;
 			case 'n' :
-				if (strcmp (optarg, "id") == 0) {
+				if (strcmp (optarg, "ctxt") == 0) {
+					write_mode |= NO_CTX;
+				} else if (strcmp (optarg, "id") == 0) {
 					write_mode |= NO_ID;
 				} else if (strcmp (optarg, "str") == 0) {
 					write_mode |= NO_STR;
@@ -465,6 +605,7 @@ main (int argc, char **argv)
 			}
 			po_write (pof, write_mode);
 		}
+		po_free (pof);
 	} else {
 		PoFile *bpof, *pof;
 		PoEntry_set *bpo_set;
@@ -474,11 +615,13 @@ main (int argc, char **argv)
 		bpo_set = po_set_create (bpof->entries);
 		pof = po_read (fn);
 		po_apply_filters (pof, ifilters);
-		bpo_set = po_set_update (bpo_set, pof->entries);
 		if (copy_msgid) {
 			po_copy_msgid (pof);
 		}
+		bpo_set = po_set_update (bpo_set, pof->entries);
 		po_write (bpof, write_mode);
+		po_free (bpof);
+		po_free (pof);
 	}
 
 	return 0;
